__version__ = "1.0.1"


# This is needed to allow Airflow to pick up specific metadata fields it needs for certain features.
def get_provider_info():
    return {
        "package-name": "apache-airflow-providers-isolation",  # Required
        "name": "Isolated Operators",  # Required
        "description": "Runtime Operator Isolation in Airflow",  # Required
        "versions": [__version__],  # Required
    }



import logging
import os
import shutil
import sys
import textwrap
from pathlib import Path
from typing import Optional
from urllib.error import HTTPError
from urllib.request import urlretrieve

import click
import sh

from isolationctl import (
    AliasedGroup,
    main_echo,
    confirm_or_exit,
    confirm_or_skip,
    CONTEXT_SETTINGS,
    DEFAULT_ON_FLAG_KWARGS,
    DEFAULT_OFF_FLAG_KWARGS,
    SH_KWARGS,
    FOLDER_OPTION_ARGS,
    FOLDER_OPTION_KWARGS,
    YES_OPTION_ARGS,
    YES_OPTION_KWARGS,
    EPILOG_KWARGS,
    add_folder_if_not_default,
    extract_kubeconfig_to_str,
    build_image,
    ENV_ARG_ARGS,
    ENV_ARG_KWARGS,
    create_registry_docker_container,
    ISOLATION_PROVIDER_PACKAGE,
    add_requirement,
    _add,
    _get,
    write_tag_to_dot_env,
    REGISTRY_CONTAINER_URI,
    EXAMPLE_ENVIRONMENT,
    KUBERNETES_CONN_KEY,
    AIRFLOW__ISOLATED_POD_OPERATOR__IMAGE_KEY,
)

log = logging.getLogger(__name__)
log.setLevel(
    os.getenv("LOG_LEVEL", logging.INFO),
)
log.addHandler(logging.StreamHandler())


@click.group(cls=AliasedGroup, context_settings=CONTEXT_SETTINGS, **EPILOG_KWARGS)
@click.version_option(package_name="apache_airflow_providers_isolation")
def cli():
    """
    ▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄
    ██▄██░▄▄█▀▄▄▀█░██░▄▄▀█▄░▄██▄██▀▄▄▀█░▄▄▀█▀▄▀█▄░▄█░█
    ██░▄█▄▄▀█░██░█░██░▀▀░██░███░▄█░██░█░██░█░█▀██░██░█
    █▄▄▄█▄▄▄██▄▄██▄▄█▄██▄██▄██▄▄▄██▄▄██▄██▄██▄███▄██▄▄
    ▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀
    """


# noinspection SpellCheckingInspection
@cli.command(**EPILOG_KWARGS)
@click.option(*YES_OPTION_ARGS, **YES_OPTION_KWARGS)
@click.option(
    "-r",
    "--registry",
    default=None,
    required=False,
    show_envvar=True,
    type=str,
    help="The registry to push environment images to "
    "such as `docker.io/path/to/my/airflow` "
    "or `aaa.dkr.ecr.us-east-2.amazonaws.com/bbb`",
)
@click.option(
    "--local-registry/--no-local-registry",
    **DEFAULT_OFF_FLAG_KWARGS,
    help=f"Shortcut to utilize a local Docker Registry container running at {REGISTRY_CONTAINER_URI}",
)
@click.option(
    "--astro/--no-astro",
    **DEFAULT_OFF_FLAG_KWARGS,
    help="Use the Astro CLI to deploy the base project, as well as environments",
)
@click.option(
    "-e",
    "--environment",
    "env",
    default=None,
    show_default=True,
    show_envvar=True,
    type=str,
    help="What environment to deploy.",
)
@click.option(*FOLDER_OPTION_ARGS, **FOLDER_OPTION_KWARGS)
def deploy(
    yes: bool, registry: Optional[str], local_registry: Optional[str], astro: bool, env: Optional[str], folder: str
):
    """Build and deploy environments (alias: d).
    \n
    Build both the parent project and either a specific or all child ENVIRONMENTs
    """
    if local_registry and registry:
        raise click.ClickException("--local-registry and --registry cannot be used simultaneously")

    if local_registry:
        registry = REGISTRY_CONTAINER_URI

    confirm_or_exit("Building parent image - all other images will derive this one...", yes)
    parent_image = build_image(method="astro-parse", get_docker_tag=True, push=False, should_log=True)
    environments = _get(env, folder, _print=False, full_path=True)
    if len(environments):
        main_echo(
            f"Found {len(environments)} {'environment' if len(environments) == 1 else 'environments'} to build..."
        )
        for _environment in environments:
            [_, _env] = _environment.split("/")
            child_tag = f"{registry}/{parent_image}/{_env}"
            if not confirm_or_skip(f"Building environment '{_environment}' from parent image '{parent_image}'...", yes):
                build_image(
                    method="docker",
                    tag=child_tag,
                    build_args={"BASE_IMAGE": parent_image},
                    push=True,
                    get_docker_tag=False,
                    should_log=True,
                    directory=_environment,
                )
                main_echo(f"Deployed environment: '{_environment}', image: '{child_tag}'!")
    else:
        raise click.ClickException("Found 0 environments to build... Exiting!")

    if astro:
        sh.astro.deploy(**SH_KWARGS)


@cli.command(**EPILOG_KWARGS)
@click.option(*FOLDER_OPTION_ARGS, **FOLDER_OPTION_KWARGS)
@click.option("--example/--no-example", **DEFAULT_ON_FLAG_KWARGS, help="Initialize with an example environment")
@click.option(
    "--local/--no-local",
    **DEFAULT_OFF_FLAG_KWARGS,
    help="Additional initialization for a local project to use with "
    "an already initialized local Kubernetes and an Astro CLI project",
)
@click.option(
    "--local-registry/--no-local-registry",
    **DEFAULT_OFF_FLAG_KWARGS,
    help="Additional initialization for a local project to use a local Docker Registry",
)
@click.option(
    "--astro/--no-astro",
    **DEFAULT_OFF_FLAG_KWARGS,
    help="Additional initialization,  create an Astro CLI project in the current directory with `astro dev init`",
)
@click.option(
    "--git/--no-git",
    **DEFAULT_OFF_FLAG_KWARGS,
    help="Additional initialization, create a git repository in the current directory with `git init`",
)
@click.option(
    "--dependency/--no-dependency",
    **DEFAULT_OFF_FLAG_KWARGS,
    help="Additional initialization, add `apache-airflow-providers-isolation` to a `requirements.txt` file. "
    "Unnecessary with `--astro`, as it's automatically added.",
)
@click.option(
    "--cicd",
    type=click.Choice(
        [
            "Codebuild",
            "DevOps",
            "Bitbucket",
            "CircleCI",
            "DroneCI",
            "Jenkins",
            "GitHub",
            "Gitlab",
            "none",
        ],
        case_sensitive=False,
    ),
    default=None,
    show_default=True,
    show_envvar=True,
    help="Additional initialization, add a CICD template file in the current project",
)
@click.option(*YES_OPTION_ARGS, **YES_OPTION_KWARGS)
@click.pass_context
def init(
    ctx,
    folder: str,
    example: bool,
    local: bool,
    local_registry: bool,
    astro: bool,
    git: bool,
    cicd: str,
    yes: bool,
    dependency: bool,
):
    """Initialize environments (alias: i).
    \n
    Initialize environments for usage with this CLI. Intended to be used with
    an existing Airflow project."""
    folder_path = Path(folder)
    if folder_path.exists():
        main_echo(f"Environments folder '{folder}' already exists, skipping...")
    else:
        confirm_or_exit(
            f"Creating an environments folder{add_folder_if_not_default(folder)} in '{os.getcwd()}'...", yes
        )
        folder_path.mkdir(parents=True, exist_ok=True)

    if example:
        main_echo("Initializing --example environment...")
        ctx.invoke(add, env=EXAMPLE_ENVIRONMENT, yes=yes, folder=folder)

        if not confirm_or_skip("Download example DAG 'isolation_provider_example_dag.py' to 'dags/'...", yes):
            example_dag = (
                "https://raw.githubusercontent.com/astronomer/apache-airflow-providers-isolation/"
                "main/isolation/example_dags/isolation_provider_example_dag.py"
            )
            try:
                Path("dags").mkdir(exist_ok=True, parents=True)
                urlretrieve(example_dag, "dags/isolation_provider_example_dag.py")
            except HTTPError as e:
                main_echo(f"Error finding example DAG: {example_dag} -- Reason:{e.reason}")

    if astro:
        requirements_txt = Path("requirements.txt")
        main_echo("Initializing --astro airflow project...")
        if shutil.which("astro") is not None:
            main_echo("astro found...")
            if not confirm_or_skip("Initializing Astro Project with astro dev init...", yes):
                sh.astro.dev.init(_in="y", _out=sys.stdout, _err=sys.stderr)
                add_requirement(requirements_txt)
        else:
            main_echo("Cannot find astro. Expecting astro to be installed. Skipping...")

    if dependency and not astro:
        requirements_txt = Path("requirements.txt")
        if not confirm_or_skip(f"Adding '{ISOLATION_PROVIDER_PACKAGE}' to '{requirements_txt.name}'...", yes):
            add_requirement(requirements_txt)

    if git:
        main_echo("Initializing --git repository...")
        if shutil.which("git") is not None:
            # noinspection PyUnresolvedReferences
            from sh.contrib import git as _git

            _git.init(_out=sys.stdout)

    if cicd:
        main_echo("Initializing --cicd templates...")
        if not Path(".git").exists():
            main_echo(
                "--cicd doesn't make sense unless this is a git repository. " "Use --git to initialize. Skipping..."
            )
        raise NotImplementedError("--cicd is not yet implemented")

    if local:
        main_echo("Initializing --local connection...")
        dot_env = Path(".env")
        if astro and local_registry:
            if shutil.which("astro") is not None:
                if not confirm_or_skip(
                    "--astro is also set: Running `astro dev parse` to fetch the initial parent image tag, "
                    "and persisting to key: 'AIRFLOW__ISOLATED_POD_OPERATOR__IMAGE' in the '.env' file...",
                    yes,
                ):
                    tag = build_image("astro-parse", should_log=False)
                    if tag:
                        write_tag_to_dot_env(REGISTRY_CONTAINER_URI, tag, dot_env)
                        main_echo(
                            f"Wrote '{AIRFLOW__ISOLATED_POD_OPERATOR__IMAGE_KEY}='{REGISTRY_CONTAINER_URI}/{tag}'"
                            " to  .env..."
                        )
                    else:
                        main_echo("Unable to find tag from `astro dev parse` output. Skipping...")
            else:
                main_echo("Cannot find astro. Expecting astro to be installed. Skipping...")

        if shutil.which("kubectl") is not None:
            main_echo("kubectl found...")
            encoded_kubeconfig = extract_kubeconfig_to_str()
            kubernetes_conn_value = (
                "kubernetes://?extra__kubernetes__namespace=default"
                f"&extra__kubernetes__kube_config={encoded_kubeconfig}"
            )
            if not dot_env.exists():
                if not confirm_or_skip(".env file not found - Creating...", yes):
                    dot_env.touch(exist_ok=True)
            else:
                main_echo(".env file found...")
            if not confirm_or_skip(
                "Writing KUBERNETES_DEFAULT Airflow Connection for local kubernetes to .env file...", yes
            ):
                dot_env.open("a").write(f"{KUBERNETES_CONN_KEY}={kubernetes_conn_value}\n")
            if not confirm_or_skip(
                "Writing AIRFLOW__ISOLATED_POD_OPERATOR__* variables for local kubernetes to .env file...", yes
            ):
                dot_env.open("a").write(
                    textwrap.dedent(
                        """
                        AIRFLOW__ISOLATED_POD_OPERATOR__KUBERNETES_CONN_ID='kubernetes_default'
                        AIRFLOW__ISOLATED_POD_OPERATOR__IMAGE_PULL_POLICY="Never"
                        """
                    )
                )
        else:
            main_echo("Cannot find kubectl. " "Expecting Kubernetes to be set up for local execution. Skipping...")

    main_echo("Initialized!")

    if local_registry:
        main_echo("Initializing --local-registry docker Image Registry...")
        if shutil.which("docker") is not None:
            if not confirm_or_skip("Creating a local docker Image Registry container on port 5000...", yes):
                create_registry_docker_container()
        else:
            main_echo("Cannot find Docker! Cannot run local Image Registry! Skipping...")


@cli.group(**EPILOG_KWARGS)
def environment():
    """Manage environments (alias: env, e). Environments are contained in folders inside an Airflow project.
    \n
    An environment inherits from a parent image (the root `Dockerfile`).
    Both images should contain all DAGs and other supporting source code.
    The child image may contain different dependencies or system packages from the parent.

    The expected setup is as follows:
    ```shell
    $ tree
    ├── Dockerfile
    ├── dags
    │  └── my_dag.py
    ├── environments
    │  ├── env_one
    │  │   ├── Dockerfile
    │  │   ├── packages.txt
    │  │   └── requirements.txt
    │  └── env_two
    │      ├── Dockerfile
    │      ├── packages.txt
    │      └── requirements.txt
    ├── packages.txt
    └── requirements.txt
    ```
    """


# noinspection PyShadowingBuiltins
@environment.command(**EPILOG_KWARGS)
@click.argument(*ENV_ARG_ARGS, **{**ENV_ARG_KWARGS, **dict(required=False)})
@click.option(*FOLDER_OPTION_ARGS, **FOLDER_OPTION_KWARGS)
def get(env: Optional[str], folder: str):
    """Get environments or specific ENVIRONMENT"""
    _get(env, folder)


# noinspection DuplicatedCode
@environment.command(**EPILOG_KWARGS)
@click.argument(*ENV_ARG_ARGS, **ENV_ARG_KWARGS)
@click.option(*YES_OPTION_ARGS, **YES_OPTION_KWARGS)
@click.option(*FOLDER_OPTION_ARGS, **FOLDER_OPTION_KWARGS)
def add(env: str, yes: bool, folder: str):
    """Add an ENVIRONMENT initialized from a template
    \n
    An environment inherits from a parent image (the root `Dockerfile`).
    Both images should contain all DAGs and other supporting source code.
    The child image may contain different dependencies or system packages from the parent.
    """
    confirm_or_exit(f"Adding environment '{env}'{add_folder_if_not_default(folder)}...", yes)
    _environment = Path(folder) / env
    if _environment.exists():
        main_echo(
            f"Environment '{env}'{add_folder_if_not_default(folder)} already exists " f"- to recreate, remove first!"
        )
    else:
        _add(_environment.name, folder)
        main_echo(f"Environment '{env}'{add_folder_if_not_default(folder)} created!")


# noinspection DuplicatedCode
@environment.command(**EPILOG_KWARGS)
@click.argument(*ENV_ARG_ARGS, **ENV_ARG_KWARGS)
@click.option(*YES_OPTION_ARGS, **YES_OPTION_KWARGS)
@click.option(*FOLDER_OPTION_ARGS, **FOLDER_OPTION_KWARGS)
def remove(env: str, yes: bool, folder: str):
    """Removes an ENVIRONMENT (which is the same as `rm -rf environments/environment`)"""
    if not confirm_or_skip(f"Removing environment '{env}'{add_folder_if_not_default(folder)}.", yes):
        _environment = Path(folder) / env
        if not _environment.exists():
            main_echo(f"Environment '{env}'{add_folder_if_not_default(folder)} does not exist!")
        else:
            shutil.rmtree(_environment)
            main_echo(f"Environment '{env}'{add_folder_if_not_default(folder)} removed!")


if __name__ == "__main__":
    cli()



import os
from datetime import datetime

from airflow import DAG
from airflow.operators.bash import BashOperator
from airflow.operators.python import PythonOperator
from airflow.providers.http.operators.http import SimpleHttpOperator

from isolation.operators.isolation import IsolatedOperator

os.environ["AIRFLOW_CONN_GITHUB"] = "http://api.github.com/https"


def print_pandas_version(arg, ds, params):
    # Note: this doesn't _need_ to be inside this function,
    # it can be in a separate file or in the top-level
    # as long as the version that Airflow has doesn't have an issue with it
    from pandas import __version__ as pandas_version

    print(f"Pandas Version: {pandas_version} \n" "And printing other stuff for fun: \n" f"{arg=}, {ds=}, {params=}")


with DAG(
    "isolation_provider_example_dag",
    schedule=None,
    start_date=datetime(1970, 1, 1),
    params={"hi": "hello"},
):
    PythonOperator(
        task_id="parent_pandas",
        python_callable=print_pandas_version,
        op_args=["Hello!"],
    )

    IsolatedOperator(
        task_id="isolated_environment_pandas",
        operator=PythonOperator,
        environment="example",
        _python_callable_qualname="dags.isolation_provider_example_dag.print_pandas_version",
        op_args=["Hello!"],
    )

    IsolatedOperator(
        task_id="echo_with_bash_operator",
        operator=BashOperator,
        environment="example",
        bash_command="echo {{ params.hi }}",
    )

    IsolatedOperator(
        task_id="use_a_connection",
        operator=SimpleHttpOperator,
        environment="example",
        endpoint="repos/apache/airflow",
        method="GET",
        http_conn_id="GITHUB",
        log_response=True,
    )



from abc import abstractmethod
from typing import Literal

from airflow.models import BaseOperator


class IsolatedOperator(BaseOperator):
    def __new__(cls, isolation_type: Literal["kubernetes"] = "kubernetes", *args, **kwargs):
        if isolation_type == "kubernetes":
            from isolation.operators.isolation_kubernetes import IsolatedKubernetesPodOperator

            return IsolatedKubernetesPodOperator(*args, **kwargs)
        else:
            raise NotImplementedError("Valid values for isolation_type are: ['kubernetes']")

    @abstractmethod
    def _set_operator_via_env(self):
        raise NotImplementedError

    @abstractmethod
    def _set_operator_args_via_env(self):
        raise NotImplementedError

    @abstractmethod
    def _set_airflow_context_via_env(self):
        raise NotImplementedError

    settable_environment_variables = {
        _set_operator_via_env.__name__: "__ISOLATED_OPERATOR_OPERATOR_QUALNAME",
        _set_operator_args_via_env.__name__: "__ISOLATED_OPERATOR_OPERATOR_ARGS",
        _set_airflow_context_via_env.__name__: "__ISOLATED_OPERATOR_AIRFLOW_CONTEXT",
    }



import json
import re
import shlex
import sys
import urllib.parse
from io import StringIO
from pathlib import Path
from typing import Literal, List, Optional, Any, Dict
import sh
import yaml
import click

MAIN_ECHO_COLOR = "blue"
SKIP_ECHO_COLOR = "white"
ISOLATION_PROVIDER_PACKAGE = "apache-airflow-providers-isolation"
DOCKERFILE = r"""
ARG BASE_IMAGE
FROM ${BASE_IMAGE}

# Install system-level packages
COPY packages.txt .
USER root
RUN if [[ -s packages.txt ]]; then \
    apt-get update && cat packages.txt | tr '\r\n' '\n' | sed -e 's/#.*//' | \
        xargs apt-get install -y --no-install-recommends \
    && apt-get clean \
    && rm -rf /var/lib/apt/lists/*; \
  fi

# Install python packages
COPY requirements.txt .
RUN if grep -Eqx 'apache-airflow\s*[=~>]{1,2}.*' requirements.txt; then \
    echo >&2 "Do not upgrade by specifying 'apache-airflow' in your requirements.txt, change the base image instead!"; \
    exit 1; \
  fi; \
  pip install --no-cache-dir --root-user-action=ignore -r requirements.txt
USER astro
"""
DEFAULT_ENVIRONMENTS_FOLDER = "environments"
EXAMPLE_ENVIRONMENT = "example"
CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])
DEFAULT_ON_FLAG_KWARGS = dict(type=bool, default=True, show_default=True, show_envvar=True)
DEFAULT_OFF_FLAG_KWARGS = dict(type=bool, default=False, show_default=True, show_envvar=True)
# SH_KWARGS = dict(_in=sys.stdin, _out=sys.stdout, _err=sys.stderr, _tee=True)
SH_KWARGS = dict(_fg=True)
FOLDER_OPTION_ARGS = ["-f", "--folder"]
FOLDER_OPTION_KWARGS = dict(
    type=click.Path(file_okay=False, dir_okay=True, readable=True, writable=True),
    default=DEFAULT_ENVIRONMENTS_FOLDER,
    show_default=True,
    show_envvar=True,
    help="Path containing Environments",
)
ENV_ARG_ARGS = ["env"]
ENV_ARG_KWARGS = dict(type=str, required=True)
YES_OPTION_ARGS = ["--yes", "-y"]
YES_OPTION_KWARGS = dict(
    type=bool, show_default=True, show_envvar=True, is_flag=True, default=False, help="Skip confirmation"
)
EPILOG_KWARGS = dict(
    epilog="Check out https://github.com/astronomer/apache-airflow-providers-isolation for more details"
)
DOCKER_TAG_PATTERN = re.compile(r".*naming to docker\.io/([\w\-/]+)[\s:].*", re.DOTALL)
REGISTRY_CONTAINER_NAME = "registry"
REGISTRY_CONTAINER_IMAGE = "registry:2"
REGISTRY_CONTAINER_URI = "localhost:5000"

# We don't *need* apache-airflow-providers-isolation[kubernetes] in child environments - but - can't hurt to add?
ADD_PROVIDER_TO_ENVIRONMENT = True
AIRFLOW__ISOLATED_POD_OPERATOR__IMAGE_KEY = "AIRFLOW__ISOLATED_POD_OPERATOR__IMAGE"
KUBERNETES_CONN_KEY = "AIRFLOW_CONN_KUBERNETES_DEFAULT"


def write_tag_to_dot_env(
    registry: str,
    image: str,
    dot_env: Path,
):
    if not dot_env.exists():
        main_echo(".env file doesn't exist - creating...")
        dot_env.touch(exist_ok=True)

    with dot_env.open(mode="a") as f:
        f.write(f"{AIRFLOW__ISOLATED_POD_OPERATOR__IMAGE_KEY}='{registry}/{image}'\n")


def get_provider_package(extras: str = "kubernetes") -> str:
    return f"{ISOLATION_PROVIDER_PACKAGE}[{extras}]"


def add_requirement(requirements_txt: Path, extras: str = "kubernetes") -> None:
    """Add apache-airflow-providers-isolation[kubernetes] to requirements.txt"""
    provider_package = get_provider_package(extras)
    if requirements_txt.exists():
        main_echo(f"'{requirements_txt.name}' file found, appending provider...")
        existing_contents = requirements_txt.read_text()
        prefix = "" if not existing_contents or existing_contents[:-2] == "\n" else "\n"
        with requirements_txt.open("a") as f:
            f.write(f"{prefix}{provider_package}\n")
    elif requirements_txt.exists() and ISOLATION_PROVIDER_PACKAGE in requirements_txt.read_text():
        main_echo(f"'{requirements_txt.name}' already contains {ISOLATION_PROVIDER_PACKAGE}! Skipping...")
    else:
        requirements_txt.touch(exist_ok=True)
        requirements_txt.write_text(f"{provider_package}\n")


def print_table(header: List[str], rows: List[List[str]]):
    longest = max(*[len(row) for row in header], *[len(col) for row in rows for col in row]) + 2
    headers = [col.ljust(longest, " ") for col in header]
    rows = [[col.ljust(longest, " ") for col in row] for row in rows]
    headers = f'\n{" | ".join(headers)}' + " |"
    separator = "-" * (len(headers) - 2) + "|"
    rows = "\n".join([" | ".join([col for col in row]) + " |" for row in rows])
    table_str = "\n".join([headers, separator, rows]) + "\n"
    sys.stdout.write(table_str)
    return table_str


def build_image_via_docker(
    tag: str,
    build_args: Optional[Dict[str, Any]] = None,
    push: bool = False,
    should_log: bool = False,
    directory: str = ".",
) -> str:
    """docker build . --tag=TAG --build-arg="BUILD_ARGS" """
    import docker

    if build_args is None:
        build_args = {}

    client = docker.from_env()

    image, logs = client.images.build(path=directory, tag=tag, buildargs=build_args)
    if should_log:
        for log in logs:
            sys.stdout.write(json.dumps(log))

    if push:
        logs = client.images.push(tag, stream=True, decode=False)
        if should_log:
            for log in logs:
                sys.stdout.write(log.decode())

    return tag


def build_image_via_astro_parse(should_log: bool = False, directory: str = ".") -> str:
    """astro dev parse, which happens to build the image as a side effect"""
    s = StringIO()
    sh.astro.dev.parse(_in=sys.stdin, _out=s, _cwd=directory, _err_to_out=True, _tee=True, _tty_size=(24, 200))
    output = s.getvalue()
    if should_log:
        sys.stdout.write(output)
    return output


def build_image_via_astro_pytest(should_log: bool = False, directory: str = ".") -> str:
    """astro dev pytest, which happens to build the image as a side effect"""
    s = StringIO()
    sh.astro.dev.pytest(_in=sys.stdin, _out=s, _cwd=directory, _err_to_out=True, _tee=True, _tty_size=(24, 500))
    output = s.getvalue()
    if should_log:
        sys.stdout.write(output)
    return output


def build_image(
    method: Literal["docker", "astro-parse", "astro-pytest"] = "astro-parse",
    tag: str = "parent",
    build_args: Optional[Dict[str, Any]] = None,
    push: bool = False,
    get_docker_tag: bool = True,
    should_log: bool = True,
    directory: str = ".",
) -> Optional[str]:
    """Build a docker image via one of multiple methods"""
    docker_tag, output = None, None
    if method == "docker":
        docker_tag = build_image_via_docker(tag, build_args, push, should_log, directory)
    elif method == "astro-parse":
        output = build_image_via_astro_parse(should_log, directory)
    elif method == "astro-pytest":
        output = build_image_via_astro_pytest(should_log, directory)
    else:
        raise RuntimeError("Methods for building the parent image are astro or docker")

    if get_docker_tag:
        if method == "docker":
            main_echo(f"Got image tag: '{docker_tag}'...")
            return docker_tag
        else:
            docker_tag = DOCKER_TAG_PATTERN.match(output)
            if docker_tag:
                docker_tag = docker_tag.group(1)
                main_echo(f"Got image tag: '{docker_tag}'...")
                return docker_tag
            else:
                raise click.ClickException(
                    f"Error! Cannot find docker_tag with regex {DOCKER_TAG_PATTERN}. " f"Unable to proceed... Exiting!"
                )


def main_echo(msg) -> None:
    click.echo(click.style(msg, fg=MAIN_ECHO_COLOR, bold=True))


def skip_echo(msg) -> None:
    click.echo(click.style(msg, fg=SKIP_ECHO_COLOR))


def confirm_or_exit(msg: str, yes: bool) -> None:
    if not yes:
        click.confirm(click.style(msg + " Continue?", fg=MAIN_ECHO_COLOR, bold=True), default=True, abort=True)
    else:
        main_echo(msg)


def confirm_or_skip(msg: str, yes: bool) -> bool:
    if not yes:
        if not click.confirm(click.style(msg + " Continue?", fg=MAIN_ECHO_COLOR, bold=True), default=True):
            skip_echo("Skipping...")
            return True
    else:
        main_echo(msg)
    return False


class AliasedGroup(click.Group):
    def get_command(self, ctx, cmd_name):
        # noinspection DuplicatedCode
        rv = click.Group.get_command(self, ctx, cmd_name)
        if rv is not None:
            return rv
        matches = [x for x in self.list_commands(ctx) if x.startswith(cmd_name)]
        if not matches:
            return None
        elif len(matches) == 1:
            return click.Group.get_command(self, ctx, matches[0])
        ctx.fail(f"Too many matches: {', '.join(sorted(matches))}")

    def resolve_command(self, ctx, args):
        # always return the full command name
        _, cmd, args = super().resolve_command(ctx, args)
        return cmd.name, cmd, args


def add_folder_if_not_default(
    folder: str, prefix=" in '", suffix="'", default: str = DEFAULT_ENVIRONMENTS_FOLDER
) -> str:
    return f"{prefix}{folder}{suffix}" if folder != default else ""


def extract_kubeconfig_to_str(replace_for_docker: bool = True) -> str:
    from io import StringIO

    s = StringIO()
    # noinspection PyUnresolvedReferences
    sh.kubectl.config.view("--raw", _out=s)
    return urllib.parse.quote(
        yaml.dump(
            yaml.safe_load(
                s.getvalue().replace("127.0.0.1", ("host.docker.internal" if replace_for_docker else "127.0.0.1"))
            )
        )
    )


def find_docker_container(name: str):
    from io import StringIO

    out = StringIO()
    # noinspection PyUnresolvedReferences
    sh.docker.ps(shlex.split(f"--all --filter='name={name}' --format=json"), _out=out)
    result = out.getvalue()
    if result:
        try:
            return json.loads(out.getvalue())
        except Exception as e:
            sys.stderr.write(str(e))
    return None


def create_registry_docker_container():
    # noinspection PyUnresolvedReferences
    try:
        if find_docker_container(REGISTRY_CONTAINER_NAME):
            main_echo("Local docker Image Registry container already created... Recreating...")
            # noinspection PyUnresolvedReferences
            sh.docker.start(REGISTRY_CONTAINER_NAME, **SH_KWARGS)
        else:
            # noinspection PyUnresolvedReferences
            sh.docker.run(
                shlex.split(
                    "-d -p 5000:5000 --restart=always " f"--name {REGISTRY_CONTAINER_NAME} {REGISTRY_CONTAINER_IMAGE}"
                ),
                **SH_KWARGS,
            )
            main_echo(f"Local docker Image Registry container created! Accessible via {REGISTRY_CONTAINER_URI} !")
    except sh.ErrorReturnCode_125:
        main_echo(
            "Container is already running, skipping... "
            "To stop and remove this container, run:\n\tdocker rm -f registry"
        )


def _add(env_name: str, folder_name: str):
    """Add an environment
    - Create the folder
    - Add a Dockerfile
    - Add a requirements.txt
    - Add a packages.txt
    """
    env = Path(env_name)
    folder = Path(folder_name)
    (folder / env).mkdir(parents=True, exist_ok=True)
    main_echo(f"Folder '{add_folder_if_not_default(folder_name, prefix='', suffix='/')}{env}/' created!")

    dockerfile = folder / env / "Dockerfile"
    dockerfile.touch()
    dockerfile.write_text(DOCKERFILE)
    main_echo(f"'{add_folder_if_not_default(folder_name, prefix='', suffix='/')}{env}/Dockerfile' added!")

    requirements_txt = folder / env / "requirements.txt"
    requirements_txt.touch()
    main_echo(f"'{add_folder_if_not_default(folder_name, prefix='', suffix='/')}{env}/requirements.txt' added!")
    if ADD_PROVIDER_TO_ENVIRONMENT:
        add_requirement(requirements_txt)

    packages_txt = folder / env / "packages.txt"
    packages_txt.touch()
    main_echo(f"'{add_folder_if_not_default(folder_name, prefix='', suffix='/')}{env}/packages.txt' added!")


def _get(
    env: Optional[str],
    env_folder: Optional[str] = DEFAULT_ENVIRONMENTS_FOLDER,
    _print: bool = True,
    full_path: bool = False,
) -> List[str]:
    header = ["Environments"]
    if env:
        envs = [
            add_folder_if_not_default(env_folder, prefix="", suffix="/", default="" if full_path else None)
            + folder.name
            for folder in Path(env_folder).iterdir()
            if folder.name == env
        ]
    else:
        envs = [
            add_folder_if_not_default(env_folder, prefix="", suffix="/", default="" if full_path else None)
            + folder.name
            for folder in Path(env_folder).iterdir()
        ]
    if len(envs) and _print:
        print_table(header, [[e] for e in envs])
    return envs



import inspect
import logging
import os
import shlex
from textwrap import dedent
from typing import Callable, Dict, Any, Tuple, List, Type, Optional, Union

from airflow.exceptions import AirflowConfigException
from airflow.providers.cncf.kubernetes.operators.pod import KubernetesPodOperator
from airflow.utils.context import Context

from isolation.operators import AIRFLOW__ISOLATED_POD_OPERATOR__IMAGE_KEY
from isolation.operators.isolation import IsolatedOperator
from isolation.util import (
    conn_template_pattern,
    conn_property_pattern,
    var_template_patterns,
    b64encode_json,
    export_to_qualname,
)

try:
    from airflow.models import BaseOperator
except AirflowConfigException:
    # Sometimes pytest gets weird here, and we are just using it for typing
    pass


# Micro-fn to inject this python into the container
def run_in_pod():
    from isolation.hooks.post_isolation import PostIsolationHook

    PostIsolationHook.run_isolated_task()


def fn_to_source_code(fn: Callable) -> str:
    return "".join(dedent(line) for line in inspect.getsourcelines(fn)[0][1:])


# noinspection GrazieInspection,SpellCheckingInspection
def _set_isolated_logger_configs(isolated_operator_kwargs: Dict[str, Any]) -> Dict[str, Any]:
    """Set some Airflow Configs related to logging for the isolated Airflow
    >>> _set_isolated_logger_configs({}) # doctest: +ELLIPSIS
    {'env_vars': {'AIRFLOW__LOGGING__COLORED_CONSOLE_LOG': 'false', 'AIRFLOW__LOGGING__LOG_FORMAT': ...
    """
    env_vars = isolated_operator_kwargs.get("env_vars", {})
    if "AIRFLOW__LOGGING__COLORED_CONSOLE_LOG" not in env_vars:
        env_vars["AIRFLOW__LOGGING__COLORED_CONSOLE_LOG"] = "false"
    if "AIRFLOW__LOGGING__LOG_FORMAT" not in env_vars:
        env_vars["AIRFLOW__LOGGING__LOG_FORMAT"] = "%(levelname)s - %(message)s"
    if "OPENLINEAGE_DISABLED" not in env_vars:
        env_vars["OPENLINEAGE_DISABLED"] = "true"
    isolated_operator_kwargs["env_vars"] = env_vars
    return isolated_operator_kwargs


def _set_default_namespace_configs(isolated_operator_kwargs: Dict[str, Any]) -> None:
    from airflow.configuration import conf

    if "namespace" not in isolated_operator_kwargs:
        isolated_operator_kwargs["namespace"] = conf.get("kubernetes", "NAMESPACE")


def _remove_un_settable_isolated_operator_kwargs(isolated_operator_kwargs: Dict[str, Any]) -> None:
    for k in IsolatedKubernetesPodOperator.isolated_operator_args:
        if k in isolated_operator_kwargs:
            logging.warning(
                "The following cannot be set in 'isolated_operator_kwargs' "
                "and must be set in IsolatedOperator or left unset - "
                f"{IsolatedKubernetesPodOperator.isolated_operator_args} . Found: {k}"
            )
            del isolated_operator_kwargs[k]


def _set_simple_templates_via_env(
    args: Tuple[Any, ...], kwargs: Dict[str, Any], isolated_operator_kwargs: Dict[str, Any]
) -> Dict[str, Any]:
    # noinspection PyTrailingSemicolon
    """Set {{var}} and {{conn}} templates, and *conn_id. Get the values and set them as env vars for the KPO"""
    from airflow.models import Connection
    from airflow.models import Variable

    env_vars = isolated_operator_kwargs.get("env_vars", {})

    def set_conns_from_values(_value):
        if not isinstance(_value, str):
            return
        results = conn_template_pattern.findall(_value)
        if results:
            for conn_id in results:
                # there will, likely, only be one here
                logging.debug(f"Passing connection {conn_id} through to " f"IsolatedOperator via ENV VAR")
                env_vars[f"AIRFLOW_CONN_{conn_id.upper()}"] = Connection.get_connection_from_secrets(
                    conn_id=conn_id
                ).get_uri()

    def set_conns_from_key_and_values(_key, _value):
        if not isinstance(_value, str) or not isinstance(_key, str):
            return
        results = conn_property_pattern.findall(f"{_key}='{_value}'")
        if results:
            for conn_id in results:
                # there will, likely, only be one here
                logging.debug(f"Passing connection {conn_id} through to " f"IsolatedOperator via ENV VAR")
                env_vars[f"AIRFLOW_CONN_{conn_id.upper()}"] = Connection.get_connection_from_secrets(
                    conn_id=conn_id
                ).get_uri()

    def set_vars_from_values(_value):
        if not isinstance(_value, str):
            return
        for results in [pattern.findall(_value) for pattern in var_template_patterns]:
            if results:
                for var_id in results:
                    logging.debug(f"Passing variable {var_id} through to " f"IsolatedOperator via ENV VAR")
                    env_vars[f"AIRFLOW_VAR_{var_id.upper()}"] = Variable.get(var_id)

    # Iterate over *args
    for i, val in enumerate(args):
        logging.debug(f"Checking args[{i}]={val} for vars and conns...")
        if val:
            # look for {{ var.xyz }}
            set_vars_from_values(val)

            # look for {{ conn.xyz }}
            set_conns_from_values(val)

    # Iterate over **kwargs
    for key, val in kwargs.items():
        logging.debug(f"Checking {key}={val} ...")
        if val:
            # look for {{ var.xyz }}
            set_vars_from_values(val)

            # look for {{ conn.xyz }}
            set_conns_from_values(val)

            if key:
                # look for *conn_id="..."
                set_conns_from_key_and_values(key, val)

    isolated_operator_kwargs["env_vars"] = env_vars
    return isolated_operator_kwargs


def _set_pod_name_configs(task_id: str, isolated_operator_kwargs: Dict[str, Any]) -> None:
    """Set the KPO Pod Name to the task_id"""
    if "name" not in isolated_operator_kwargs:
        isolated_operator_kwargs["name"] = task_id


def _set_airflow_context_via_env(context: Context, env_vars: List["V1EnvVar"]) -> List["V1EnvVar"]:  # noqa: F821
    """Serialize the Airflow Context
    >>> _set_airflow_context_via_env(Context({"ds": "foo", "params": {"foo": "bar"}}), [])
    [{'name': '__ISOLATED_OPERATOR_AIRFLOW_CONTEXT',
     'value': 'eyJkcyI6ICJmb28iLCAicGFyYW1zIjogeyJmb28iOiAiYmFyIn19',
     'value_from': None}]
    """
    from airflow.providers.cncf.kubernetes.backcompat.backwards_compat_converters import convert_env_vars

    key = IsolatedOperator.settable_environment_variables[_set_airflow_context_via_env.__name__]
    env_vars.extend(convert_env_vars({key: b64encode_json(dict(**context))}))
    return env_vars


def _set_operator_via_env(
    operator: Union[str, Type["BaseOperator"]],
    isolated_operator_kwargs: Dict[str, Any],
    calling_dag_filename: Optional[str] = None,
) -> Dict[str, Any]:
    """Turn an Operator a direct qualified name reference, so it can get resurrected on the other side
    >>> from airflow.operators.bash import BashOperator; kw = {}; _set_operator_via_env(BashOperator, kw)
    {'env_vars': {'__ISOLATED_OPERATOR_OPERATOR_QUALNAME': 'airflow.operators.bash.BashOperator'}}
    >>> kw = {}; _set_operator_via_env('airflow.operators.bash.BashOperator', kw)
    {'env_vars': {'__ISOLATED_OPERATOR_OPERATOR_QUALNAME': 'airflow.operators.bash.BashOperator'}}
    """
    env_vars = isolated_operator_kwargs.get("env_vars", {})
    key = IsolatedOperator.settable_environment_variables[_set_operator_via_env.__name__]
    env_vars[key] = (
        operator
        if isinstance(operator, str)
        else export_to_qualname(operator, validate=True, calling_dag_filename=calling_dag_filename)
    )
    isolated_operator_kwargs["env_vars"] = env_vars
    return isolated_operator_kwargs


def _set_callable_qualname(d: Dict[str, Any], calling_dag_filename: Optional[str] = None) -> Dict[str, Any]:
    """Turn xyz_callable=fn into _xyz_callable_qualname=qualname. It gets turned back on the other side
    >>> _set_callable_qualname({}) # empty
    {}
    >>> _set_callable_qualname({"python_callable": _set_callable_qualname})  # happy path
    {'_python_callable_qualname': 'isolation.operators.isolation_kubernetes._set_callable_qualname'}
    >>> _set_callable_qualname({"python_callable": "foo"})  # no change
    {'python_callable': 'foo'}
    """
    keys = list(d.keys())
    for key in keys:
        value = d[key]
        if isinstance(value, Callable):
            new_key = f"_{key}_qualname"
            new_value = export_to_qualname(value, validate=False, calling_dag_filename=calling_dag_filename)
            d[new_key] = new_value
            del d[key]
    return d


def _convert_args(
    args: Tuple[Any, ...], kwargs: Dict[str, Any], calling_dag_filename: Optional[str] = None
) -> Dict[str, Any]:
    """Pack the args and kwargs as json, convert any Callables to qualified names
    >>> _convert_args((), {})
    {'args': (), 'kwargs': {}}
    """
    kwargs = _set_callable_qualname(kwargs, calling_dag_filename)
    return {"args": args, "kwargs": kwargs}


def _set_operator_args_via_env(
    args: Tuple[Any, ...],
    kwargs: Dict[str, Any],
    isolated_operator_kwargs: Dict[str, Any],
    calling_dag_filename: Optional[str] = None,
):
    """Set args and kwargs that were given for the operator to __ISOLATED_OPERATOR_OPERATOR_ARGS
     1) b64 and json encode them
     2) if it is xyz_callable, we rename it to _xyz_callable_qualname and input the qualified name of the fn
    >>> _set_operator_args_via_env((), {}, {})
    {'env_vars': {'__ISOLATED_OPERATOR_OPERATOR_ARGS': 'eyJhcmdzIjogW10sICJrd2FyZ3MiOiB7fX0='}}
    >>> _set_operator_args_via_env(('foo',), {"bar_callable": print}, {"existing_kwargs": []}) # doctest: +ELLIPSIS
    {'existing_kwargs': [], 'env_vars': {'__ISOLATED_OPERATOR_OPERATOR_ARGS': ...
    """
    key = IsolatedOperator.settable_environment_variables[_set_operator_args_via_env.__name__]
    env_vars = isolated_operator_kwargs.get("env_vars", {})
    env_vars[key] = b64encode_json(_convert_args(args, kwargs, calling_dag_filename))
    isolated_operator_kwargs["env_vars"] = env_vars
    return isolated_operator_kwargs


def _set_deferrable(
    kwargs: Dict[str, Any], isolated_operator_kwargs: Dict[str, Any]
) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    """If the 'deferrable' arg is set, ensure if gets set on the KPO only
    >>> _set_deferrable({"deferrable": False}, {})
    ({}, {'deferrable': False})
    """
    if "deferrable" in kwargs:
        value = kwargs["deferrable"]
        del kwargs["deferrable"]
        isolated_operator_kwargs["deferrable"] = value
    return kwargs, isolated_operator_kwargs


# noinspection PyTrailingSemicolon
def _set_kpo_default_args_from_env(isolated_operator_kwargs: Dict[str, Any]) -> Dict[str, Any]:
    """Scan env for any key like AIRFLOW__ISOLATED_POD_OPERATOR__XYZ and default it
    >>> os.environ["AIRFLOW__ISOLATED_POD_OPERATOR__KUBERNETES_CONN_ID"]="FOO"; _set_kpo_default_args_from_env(
    ...     {}
    ... ); del os.environ["AIRFLOW__ISOLATED_POD_OPERATOR__KUBERNETES_CONN_ID"]
    {'kubernetes_conn_id': 'FOO'}
    >>> os.environ["AIRFLOW__ISOLATED_POD_OPERATOR__IMAGE"]="FOO"; _set_kpo_default_args_from_env({})
    {}
    """
    prefix = "AIRFLOW__ISOLATED_POD_OPERATOR__"
    for key, value in os.environ.items():
        if prefix in key:
            key = key.replace(prefix, "").lower()
            # If it's not in the args list, and it isn't already set
            if key not in IsolatedKubernetesPodOperator.isolated_operator_args and key not in isolated_operator_kwargs:
                isolated_operator_kwargs[key] = value
    return isolated_operator_kwargs


def _derive_image(image: Optional[str], default_image_from_env: Optional[str], environment: Optional[str]) -> str:
    """Take the different sources of "image" and "environment" and derive a final image
    1) return image/env, if given
    2) return image, if env not given
    3) return default_image/env, if given
    4) error otherwise

    >>> _derive_image(None,None,None)  # Give nothing, get an error
    Traceback (most recent call last):
        ...
    RuntimeError: Image must be set via the 'image' argument or the 'AIRFLOW__ISOLATED_POD_OPERATOR__IMAGE' env var
    >>> _derive_image("foo",None,None)  # Give an image, get the image
    'foo'
    >>> _derive_image("foo",None,"bar")
    'foo/bar'
    >>> _derive_image("foo","baz","bar")
    'foo/bar'
    >>> _derive_image(None,"baz","bar")
    'baz/bar'
    >>> _derive_image(None,None,"bar")
    Traceback (most recent call last):
        ...
    RuntimeError: Image must be set via the 'image' argument or the 'AIRFLOW__ISOLATED_POD_OPERATOR__IMAGE' env var
    >>> _derive_image(None,"baz",None)
    Traceback (most recent call last):
        ...
    RuntimeError: The 'AIRFLOW__ISOLATED_POD_OPERATOR__IMAGE' env var must be paired with the 'environment' argument
    """
    # Prefer image to the env var default
    if image:
        # if that's all we have - return it
        if not environment:
            return image

        # If we have image/environment - return it like that
        else:
            return f"{image}/{environment}"
    elif default_image_from_env:
        # if that's all we have - error
        if not environment:
            raise RuntimeError(
                f"The '{AIRFLOW__ISOLATED_POD_OPERATOR__IMAGE_KEY}' env var "
                f"must be paired with the 'environment' argument"
            )

        # If we have image/environment - return it like that
        else:
            return f"{default_image_from_env}/{environment}"

    # Both image sources are empty - error
    else:
        raise RuntimeError(
            f"Image must be set via the 'image' argument or the '{AIRFLOW__ISOLATED_POD_OPERATOR__IMAGE_KEY}' env var"
        )


# noinspection GrazieInspection,SpellCheckingInspection
def _set_post_isolation_flag_via_env(isolated_operator_kwargs: Dict[str, Any]) -> Dict[str, Any]:
    """Set the __ISOLATED_OPERATOR_POST_ISOLATION flag via the env, to prevent attempting to (re)create
    the IsolatedOperator on the other side, for example if you are loading a function from the DAG itself"""
    env_vars = isolated_operator_kwargs.get("env_vars", {})
    env_vars["__ISOLATED_OPERATOR_POST_ISOLATION"] = "True"
    isolated_operator_kwargs["env_vars"] = env_vars
    return isolated_operator_kwargs


class IsolatedKubernetesPodOperator(KubernetesPodOperator):
    """IsolatedKubernetesPodOperator - run an IsolatedOperator powered by the KubernetesPodOperator
    :param task_id str: Task id, as normal
    :param operator Type[BaseOperator]: Operator to run, e.g. operator=BashOperator.
        :class:`airflow.models.base_operator.BaseOperator`
    :param image Optional[str]: full name of image to run, without a tag, e.g. docker.io/organization/project/airflow
        default to env var AIRFLOW__ISOLATED_POD_OPERATOR__IMAGE
    :var 'AIRFLOW__ISOLATED_POD_OPERATOR__IMAGE' is retrieved from the env if set, as a default
    :param environment Optional[str]: the name of the environment which gets appended to the image.
        Just the image is used, if not given
    :param isolated_operator_kwargs Optional[Dict[str, Any]]: kwargs passed directly to
        the :class:`airflow.providers.cncf.kubernetes.operators.pod.KubernetesPodOperator`
    :param args List[str]: args for the operator
    :param kwargs Dict[str, Any]: kwargs for the operator, e.g. bash_command="echo hi"

    :Example:

    ```
    def print_pandas_version(arg, ds, params):
        from pandas import __version__ as pandas_version

        print(f"Pandas Version: {pandas_version} \n" "And printing other stuff for fun: \n" f"{arg=}, {ds=}, {params=}")

    IsolatedOperator(
        task_id="isolated_environment_pandas",
        operator=PythonOperator,
        image="localhost:5000/my-airflow-project_36d6b4/airflow",
        environment="data-team",
        python_callable=print_pandas_version,
        op_args=["Hello!"]
    )
    ```

    .. seealso:: :class:`isolation.operators.isolation.IsolatedOperator`
    """

    isolated_operator_args = [
        "task_id",
        "image",
        "cmds",
        "arguments",
        "log_events_on_failure",
    ]

    def __init__(
        self,
        task_id: str,
        operator: Type["BaseOperator"],
        image: Optional[str] = None,
        environment: Optional[str] = None,
        isolated_operator_kwargs: Optional[Dict[str, Any]] = None,
        *args,
        **kwargs,
    ):
        if "__ISOLATED_OPERATOR_POST_ISOLATION" in os.environ and os.getenv("__ISOLATED_OPERATOR_POST_ISOLATION"):
            super().__init__(task_id=task_id)
        else:
            self._args = args
            self._kwargs = kwargs
            self.isolated_operator_kwargs = isolated_operator_kwargs or {}
            self.operator = operator
            self.environment = environment
            self.image = image
            module = inspect.getmodule(inspect.stack()[1][0])
            self.calling_dag_filename = module.__file__ if module else None

            _remove_un_settable_isolated_operator_kwargs(self.isolated_operator_kwargs)
            _set_isolated_logger_configs(self.isolated_operator_kwargs)
            _set_default_namespace_configs(self.isolated_operator_kwargs)
            _set_pod_name_configs(task_id, self.isolated_operator_kwargs)
            _set_deferrable(self._kwargs, self.isolated_operator_kwargs)
            _set_simple_templates_via_env(self._args, self._kwargs, self.isolated_operator_kwargs)
            _set_operator_via_env(self.operator, self.isolated_operator_kwargs, self.calling_dag_filename)
            _set_operator_args_via_env(
                self._args, self._kwargs, self.isolated_operator_kwargs, self.calling_dag_filename
            )
            _set_kpo_default_args_from_env(self.isolated_operator_kwargs)
            default_image_from_env = os.getenv(AIRFLOW__ISOLATED_POD_OPERATOR__IMAGE_KEY)
            image = _derive_image(self.image, default_image_from_env, self.environment)
            _set_post_isolation_flag_via_env(self.isolated_operator_kwargs)

            # noinspection SpellCheckingInspection
            super().__init__(
                task_id=task_id,
                image=image,
                cmds=shlex.split("/bin/bash -euxc"),
                arguments=[f'python -c "{fn_to_source_code(run_in_pod)}"'],
                log_events_on_failure=True,
                **self.isolated_operator_kwargs,
            )

    def execute(self, context: Context):
        _set_airflow_context_via_env(context, self.env_vars)
        super().execute(context)



import inspect
import itertools
from copy import copy
from pathlib import Path
from typing import Any, Dict, Type, Tuple, List, Optional

from isolation.util import (
    b64decode_json,
    getenv_or_raise,
    get_and_check_airflow_version,
    validate_operator_is_operator,
    import_from_qualname,
    qualname_param_pattern,
)

XCOM_FILE = "/airflow/xcom/return.json"


def _verify_kwargs(kwargs: Dict[str, Any]):
    """Override or set required kwargs
    1) Set/override the task_id
    2) Set/override do_xcom_push to False

    >>> d = {}; _verify_kwargs(d); d
    {'task_id': '_ISOLATED', 'do_xcom_push': False}
    >>> d = {"foo": "bar"}; _verify_kwargs(d); d
    {'foo': 'bar', 'task_id': '_ISOLATED', 'do_xcom_push': False}
    """
    kwargs["task_id"] = "_ISOLATED"
    kwargs["do_xcom_push"] = False


def _get_operator_args_via_env(operator_args: Optional[str] = None) -> Tuple[List[Any], Dict[str, Any]]:
    # noinspection PyUnresolvedReferences, PyProtectedMember
    """Return *args and **kwargs for the operator,
    encoded as {"args":args,"kwargs":kwargs} as a b64 encoded json

    :raises: RuntimeError if extracted args is not a list
    :raises: RuntimeError if extracted kwargs is not a dict
    >>> from isolation.util import b64encode_json; import datetime, base64, json, os
    >>> _get_operator_args_via_env()  # Unset
    Traceback (most recent call last):
        ...
    RuntimeError: __ISOLATED_OPERATOR_OPERATOR_ARGS must be set!
    >>> _get_operator_args_via_env("")   # Empty
    ([], {})
    >>> _get_operator_args_via_env(
    ...     b64encode_json({"args":["foo", 1], "kwargs":{"a": 1, "b": 2, "c": datetime.datetime(1970, 1, 1)}})
    ... )   # Happy Path
    (['foo', 1], {'a': 1, 'b': 2, 'c': '1970-01-01 00:00:00'})
    >>> _get_operator_args_via_env(
    ...     b64encode_json({"args": [], "kwargs": {}})
    ... )   # Happy & Empty
    ([], {})
    """
    key = PostIsolationHook.required_environment_variables[_get_operator_args_via_env.__name__]
    encoded_args_json = getenv_or_raise(key) if operator_args is None else operator_args
    if not encoded_args_json:
        return [], {}
    args_json = b64decode_json(encoded_args_json)
    args, kwargs = args_json["args"], args_json["kwargs"]
    if args is None:
        args = []
    if not isinstance(args, list):
        raise RuntimeError(f"args is not a list! Got: {args} . Cannot proceed")

    if kwargs is None:
        kwargs = {}
    if not isinstance(kwargs, dict):
        raise RuntimeError(f"kwargs is not a list! Got: {kwargs} . Cannot proceed")

    return args, kwargs


def _get_context_via_env(context: Optional[str] = None) -> "Context":  # noqa: F821
    # noinspection PyUnresolvedReferences,PyTrailingSemicolon,SpellCheckingInspection
    """Retrieve a json-serialized version of a subset
    of the Airflow Context object and recreate
    >>> from isolation.util import b64encode_json
    >>> r = _get_context_via_env(
    ...     "eyJkcyI6ICIxOTcwLTAxLTAxVDAwOjAwOjAwIiwgInBhcmFtcyI6IHsiZm9vIjogImJhciJ9fQ=="  # pragma: allowlist secret
    ... ); r['params'], type(r['ds'])
    ({'foo': 'bar'}, <class 'pendulum.datetime.DateTime'>)
    >>> r = _get_context_via_env(b64encode_json({})); r['params'], type(r['ds']) # Empty
    ({}, <class 'pendulum.datetime.DateTime'>)
    >>> r = _get_context_via_env(''); r['params'], type(r['ds'])
    ({}, <class 'pendulum.datetime.DateTime'>)
    >>> r = _get_context_via_env(b64encode_json({"ds": "foo"})) # bad de-ser data
    Traceback (most recent call last):
        ...
    pendulum.parsing.exceptions.ParserError: Unable to parse string [foo]
    >>> r = _get_context_via_env(
    ...     b64encode_json({"params": {"foo": "bar"}})
    ... ); r['params'], type(r['params'])  # good params data
    ({'foo': 'bar'}, <class 'airflow.models.param.ParamsDict'>)
    """
    from airflow.utils.context import Context
    import pendulum
    from airflow.models.param import ParamsDict

    key = PostIsolationHook.required_environment_variables[_get_context_via_env.__name__]
    encoded_context_json = getenv_or_raise(key) if context is None else context
    context_json = b64decode_json(encoded_context_json) if encoded_context_json != "" else {}
    if not isinstance(context_json, dict):
        raise RuntimeError(f"{key} did not decode to JSON - Value: {encoded_context_json}, Decoded: {context_json}")

    required_keys = ["ds", "params"]
    context_serialization_mappings = {
        # Key:  (default-fn,  mapping-fn)
        "params": (dict, ParamsDict),
        "ds": (pendulum.now, pendulum.parse),
    }
    # we check over both what we got AND the required ones
    for k in set(itertools.chain(context_json.keys(), required_keys)):
        if k in context_json:  # if it's one of the ones we got
            if k in context_serialization_mappings:  # and if we have a mapping
                context_json[k] = context_serialization_mappings[k][1](context_json[k])  # map it
            else:  # else we got it, but
                pass  # it's fine as-is
        else:  # if it's required, wasn't given, and needs to be defaulted
            context_json[k] = context_serialization_mappings[k][0]()  # default it
    return Context(**context_json)


def _get_operator_via_env(operator_qualname: Optional[str] = None) -> Type["BaseOperator"]:  # noqa: F821
    """Instantiate the operator via the qualified name saved in __ISOLATED_OPERATOR_OPERATOR_QUALNAME
    >>> _get_operator_via_env("datetime.datetime")
    <class 'datetime.datetime'>
    """
    # Sanity check - __ISOLATED_OPERATOR_OPERATOR_QUALNAME=a.b.c.d.MyOperator
    key = PostIsolationHook.required_environment_variables[_get_operator_via_env.__name__]
    operator_qualname = getenv_or_raise(key) if operator_qualname is None else operator_qualname
    # split out - a.b.c.d, MyOperator
    return import_from_qualname(operator_qualname)


def _patch_task_dependencies(task) -> None:
    """We overwrite the method to avoid the DB call"""

    def blackhole() -> iter:
        return iter([])

    task.iter_mapped_dependants = blackhole


def _patch_post_execute_for_xcom(task, kwargs: Optional[Dict[str, Any]] = None) -> None:
    """Patch the 'task.post_execute' method - write a xcom then run the original"""
    original_fn = task.post_execute
    do_xcom_push = kwargs["do_xcom_push"] if kwargs is not None and "do_xcom_push" in kwargs else False

    def new_fn(context, result):
        if result and do_xcom_push:
            xcom_path = Path(XCOM_FILE)
            try:
                xcom_path.parent.mkdir(parents=True, exist_ok=True)
                with xcom_path.open("w") as f:
                    f.write(result)
            except PermissionError as e:
                raise RuntimeError("Unable to write XCOM") from e
        return original_fn(context, result)

    task.post_execute = new_fn


def _patch_clear_xcom_data(task_instance) -> None:
    """Patch the 'task_instance.clear_xcom_data' method - just respond immediately, skip DB"""

    def blackhole() -> None:
        return

    task_instance.clear_xcom_data = blackhole


def _get_callable_qualname(key: str) -> Optional[str]:
    """We should be getting xyz_callable as _xyz_callable_qualname - find and return if so
    >>> _get_callable_qualname("_xyz_qualname")
    'xyz'
    >>> _get_callable_qualname("_python_callable_qualname")
    'python_callable'
    >>> _get_callable_qualname("definitely_not_python_callable")
    >>> _get_callable_qualname("bash_command")
    """
    maybe_match = qualname_param_pattern.match(key)
    return maybe_match.group(1) if maybe_match is not None else None


def _transform_args(args: List[Any], kwargs: Dict[str, Any]) -> Tuple[List[Any], Dict[str, Any]]:
    """
    Look for anything like '_xyz_qualname' or '_python_callable_qualname' in the kwargs, remap it to a fn
    >>> _transform_args([], {
    ...     "_python_callable_qualname": "tests.resources.fn.direct.print_magic_string_test_direct"
    ... })  # doctest: +ELLIPSIS
    ([], {'python_callable': <function print_magic_string_test_direct at ...
    """
    dict_keys = copy(list(kwargs.keys()))
    for key in dict_keys:
        maybe_qualname_key = _get_callable_qualname(key)
        if maybe_qualname_key:
            # save whatever the callable is
            qualname = kwargs[key]

            # delete "_python_callable_qualname"
            del kwargs[key]

            # set "python_callable"
            kwargs[maybe_qualname_key] = import_from_qualname(qualname)
    return args, kwargs


class PostIsolationHook:
    required_environment_variables = {
        _get_operator_via_env.__name__: "__ISOLATED_OPERATOR_OPERATOR_QUALNAME",
        _get_operator_args_via_env.__name__: "__ISOLATED_OPERATOR_OPERATOR_ARGS",
        _get_context_via_env.__name__: "__ISOLATED_OPERATOR_AIRFLOW_CONTEXT",
    }

    # noinspection PyMethodMayBeStatic
    @classmethod
    def run_isolated_task(cls) -> None:
        """Run an Isolated Airflow Task inside a Container
        1) get operator from ENV VAR - e.g. __ISOLATED_OPERATOR_OPERATOR_QUALNAME=a.b.c.d.MyOperator
        2) get args/kwargs from ENV VAR - e.g. __ISOLATED_OPERATOR_OPERATOR_ARGS=<b64 {"args": [], "kwargs": {}}
        3) get Airflow Context from ENV VAR e.g. __ISOLATED_OPERATOR_AIRFLOW_CONTEXT=<b64 {"ds": "1970-01-01", ...}
        4) Execute the Operator with args/kwargs, hijack task.post_execute to write the XCOM
        """
        from airflow.models.taskinstance import set_current_context, TaskInstance

        # Do some sanity checks - check that we are using AF2.x
        # returns the AF major.minor version if we need to check compatability later
        get_and_check_airflow_version()

        operator = _get_operator_via_env()
        validate_operator_is_operator(operator)

        args, kwargs = _get_operator_args_via_env()
        _verify_kwargs(kwargs)
        _transform_args(args, kwargs)
        task = operator(*args, **kwargs)
        _patch_task_dependencies(task)
        _patch_post_execute_for_xcom(task, kwargs)

        context = _get_context_via_env()
        with set_current_context(context):
            # noinspection SpellCheckingInspection
            required_task_args = {"run_id": "RUNID", "state": "running"}
            ti = TaskInstance(task=task, **required_task_args)
            _patch_clear_xcom_data(ti)

            # for different airflow versions - if these fields exist, set them
            _kwarg_to_default = {"test_mode": True, "session": None}

            # noinspection PyProtectedMember
            ti._execute_task_with_callbacks(
                context,
                **{
                    _kwarg: default
                    for _kwarg, default in _kwarg_to_default.items()
                    if _kwarg in inspect.signature(ti._execute_task_with_callbacks).parameters
                },
            )



import base64
import binascii
import hashlib
import inspect
import json
import os
import re
import sys
from importlib import import_module
from pathlib import Path
from typing import Dict, Any, Optional, Type, Union, Callable


def b64encode_json(d: Dict[str, Any]) -> str:
    """helper utility to encode a dict to a b64 string

    :raises: Exception if input is not JSON serializable
    >>> b64encode_json({"foo": "bar"})
    'eyJmb28iOiAiYmFyIn0='
    """
    try:
        return base64.b64encode(json.dumps(d, default=str).encode()).decode()
    except TypeError as e:
        raise RuntimeError(f"Unable to serialize! Given: {d}") from e


def b64decode_json(s: str) -> Dict[str, Any]:
    """helper utility to decode a b64 to a json dict

    :raises: RuntimeError if input is not b64
    :raises: RuntimeError if input is not valid JSON
    >>> b64decode_json("InsnZHMnOiAnZm9vJywgJ3BhcmFtcyc6IHsnZm9vJzogJ2Jhcid9fSI=")  # pragma: allowlist secret
    "{'ds': 'foo', 'params': {'foo': 'bar'}}"
    >>> b64decode_json('eyJmb28iOiAiYmFyIn0=')
    {'foo': 'bar'}
    >>> b64decode_json("garbage")   # Garbage - not b64
    Traceback (most recent call last):
        ...
    RuntimeError: Input is not encoded correctly as base64! Got: garbage . Cannot proceed!
    >>> b64decode_json("{{}") # Garbage - not json
    Traceback (most recent call last):
        ...
    RuntimeError: Input is not encoded correctly as json! Got: b'' from '{{}'. Cannot proceed!
    """
    try:
        decoded_j = base64.b64decode(s)
        return json.loads(decoded_j)
    except binascii.Error as e:
        raise RuntimeError(
            f"Input is not encoded correctly as base64! " f"Got: {s} . Cannot proceed!",
        ) from e
    except json.decoder.JSONDecodeError as e:
        # noinspection PyUnboundLocalVariable
        raise RuntimeError(
            f"Input is not encoded correctly as json! " f"Got: {decoded_j} from '{s}'. Cannot proceed!",
        ) from e


def getenv_or_raise(key):
    """Get the env key or throw an error
    :raises: RuntimeError if key is not set
    """
    if key not in os.environ:
        raise RuntimeError(f"{key} must be set!")
    return os.getenv(key)


def get_isolated_operator_env(
    op_qualname: str,
    kwargs: Optional[Dict[str, Any]] = None,
    context: Optional[Dict[str, Any]] = None,
):
    """Returns a dict that gets passed to the Docker Container as it's `environment`"""
    return {
        "__ISOLATED_OPERATOR_OPERATOR_QUALNAME": op_qualname,
        "__ISOLATED_OPERATOR_OPERATOR_ARGS": b64encode_json({"args": [], "kwargs": kwargs}) if kwargs else "",
        "__ISOLATED_OPERATOR_AIRFLOW_CONTEXT": b64encode_json(context) if context else "",
    }


# noinspection RegExpAnonymousGroup
var_template_patterns = [
    re.compile(r"{{\s*var.value.([a-zA-Z-_]+)\s*}}"),  # "{{ var.value.<> }}"
    re.compile(r"{{\s*var.json.([a-zA-Z-_]+)\s*}}"),  # "{{ var.json.<> }}"
]
# noinspection RegExpAnonymousGroup
var_object_pattern = re.compile(r"""Variable.get[(]["']([a-zA-Z-_]+)["'][)]""")  # "Variable.get(<>)"
# noinspection RegExpAnonymousGroup
conn_property_pattern = re.compile(r"""(?=\w*_)conn_id=["']([a-zA-Z-_]+)["']""")  # "conn_id=<>"
# noinspection RegExpAnonymousGroup
conn_template_pattern = re.compile(r"[{]{2}\s*conn[.]([a-zA-Z-_]+)[.]?")  # "{{ conn.<> }}"
# noinspection RegExpAnonymousGroup
unusual_prefix_pattern = re.compile(r"unusual_prefix_(\w+)_(.+)")
# noinspection RegExpAnonymousGroup
qualname_param_pattern = re.compile(r"[_](.*)(?=_qualname)")


def get_and_check_airflow_version() -> float:
    """Normalize the Airflow Version, Steps:
    1) normalize 2.2.5+astro.6 to 2.2.5
    2) normalize 2.2.5 to 2.2
    4) returns 2.x

    :raises: RuntimeError if the version is not Airflow 2.x
    """
    from airflow import __version__

    [version, _] = __version__.rsplit("+", 1)
    [major_minor, _] = version.rsplit(".", 1)
    major, minor = major_minor.split(".")
    af_version = float(major_minor)
    if int(major) != 2:
        raise RuntimeError("PostIsolationHook only works with Airflow 2.x!")
    return af_version


def validate_operator_is_operator(operator: Type["BaseOperator"]) -> None:  # noqa: F821
    """make sure the "operator" we were given is an Operator, e.g. BashOperator
    :raises: RuntimeError - if operator is not a subclass of BaseOperator

    >>> from airflow.operators.bash import BashOperator; validate_operator_is_operator(BashOperator)  # happy path
    >>> class MyOperator(BashOperator):
    ...    pass
    >>> validate_operator_is_operator(MyOperator)  # happy path with a custom operator
    >>> validate_operator_is_operator(dict) # non-operators don't parse
    Traceback (most recent call last):
    ...
    RuntimeError: <class 'dict'> must be a subclass of <class 'airflow.models.baseoperator.BaseOperator'>
    """
    from airflow.models import BaseOperator

    if not issubclass(operator, BaseOperator):
        raise RuntimeError(f"{operator} must be a subclass of {BaseOperator}")


def fix_unusual_prefix_serialized_dag_issue(qualname, calling_filepath: Optional[str] = None):
    """
    Issue with airflow dag serialization stuff modifying module to 'unusual_prefix_xyzabc_file' if file is dags/file.py
    https://github.com/apache/airflow/blob/88b274c95b212b541ba19918880ae425856212be/airflow/models/dagbag.py#L287
    "unusual_prefix_cc53bfb719e11e2b18b1d66382f5047c2461068c_isolation_provider_example_dag"  # pragma: allowlist secret
                    cc53bfb719e11e2b18b1d66382f5047c2461068c
                    /usr/local/airflow/dags/isolation_provider_example_dag.py
    """
    # skip this code - at the moment
    if False and "unusual_prefix" in qualname:
        [module, name] = qualname.rsplit(".", 1)
        matches = unusual_prefix_pattern.match(module).groups()
        if calling_filepath and matches:
            _hash, file = matches
            files = list(Path(calling_filepath).rglob("*.[pP][yY]"))
            file_results = [path for path in files if file in path.name]
            if len(file_results):
                hash_results = [
                    result
                    for result in file_results
                    if hashlib.sha1(str(result.resolve()).encode("utf-8")).hexdigest() == _hash
                ]
                if len(hash_results):
                    if len(hash_results) > 1:
                        print(
                            f"Attempting to unwind 'unusual_prefix_*'... found {len(hash_results)} files "
                            f"with SHA1 {_hash} recursively in {calling_filepath} of {len(files)} files... "
                            "Using the first..."
                        )

                    print(f"Attempting to unwind 'unusual_prefix_*'... importing {hash_results[0].parent}...")
                    sys.path.append(str(hash_results[0].parent))
                    mod = import_module(file)
                    print(f"Attempting to unwind 'unusual_prefix_*'... {mod} {mod.__loader__}")
                    obj = getattr(mod, name)
                    print(f"Attempting to unwind 'unusual_prefix_*'... {obj}")
                    return export_to_qualname(obj, validate=False, check_unusual=False)
                print(
                    f"Attempting to unwind 'unusual_prefix_*'... "
                    f"unable to find file with SHA1 {_hash} recursively in {calling_filepath} of {len(files)} files"
                )
            print(
                f"Attempting to unwind 'unusual_prefix_*'... "
                f"unable to find file {file} recursively from {calling_filepath} in {len(files)} files"
            )
        print(
            f"Attempting to unwind 'unusual_prefix_*'... Either unable to extract hash and file from {qualname}, "
            f"or unable to get calling file: {calling_filepath}, skipping..."
        )
    return qualname


def import_from_qualname(qualname) -> Type["BaseOperator"]:  # noqa: F821
    """Turn a.b.c.d.MyOperator into the actual python version
    Steps:
    1) split out - a.b.c.d, MyOperator from a.b.c.d.MyOperator
    2) import a.b.c.d
    3) return a.b.c.d.MyOperator
    """
    qualname = fix_unusual_prefix_serialized_dag_issue(qualname)
    [module, name] = qualname.rsplit(".", 1)
    imported_module = import_module(module)
    return getattr(imported_module, name)


def export_to_qualname(
    thing: Union[Callable, Type["BaseOperator"]],  # noqa: F821
    validate: Optional[bool] = True,
    calling_dag_filename: Optional[str] = None,
    check_unusual: Optional[bool] = True,
) -> str:
    """Turn an Operator into it's qualified name
    e.g. BashOperator -> 'airflow.operators.bash.BashOperator'
    :raises: RuntimeError if Operator does not inherit from BaseOperator

    >>> from airflow.operators.bash import BashOperator; export_to_qualname(thing=BashOperator)
    'airflow.operators.bash.BashOperator'
    """
    if validate:
        validate_operator_is_operator(thing)
    qualname = f"{inspect.getmodule(thing).__name__}.{thing.__name__}"
    if check_unusual:
        qualname = fix_unusual_prefix_serialized_dag_issue(qualname, calling_dag_filename)
    return qualname
